#include "communication.h"
#include <string.h>
#include "queue.h"
#include "test.h"

struct Queue *queue;
struct UARTDriver *uart;

int mock_uart_read(struct UARTDriver *self, unsigned char *c) {
  return queue_remove(queue, c);
}

struct UARTDriver * mock_uart_malloc() {
  struct UARTDriver *uart = malloc(sizeof(struct UARTDriver));
  uart->read = mock_uart_read;
  return uart;
}

void queue_add_corruption(struct Queue *q, int num) {
  int val = 0xAA;
  int i;
  for (i = 0; i < num; i += 1) {
    val = (val << 1) ^ i;
    if (val == 0xBE) queue_add(q, 0);
    else queue_add(q, val);
  }
}

void queue_add_short(struct Queue *q, short val) {
  queue_add(q, val & 0x00FF);
  queue_add(q, val >> 8);
}

// Return checksum
unsigned char queue_add_packet(struct Queue *q,
		unsigned short type,
		unsigned short id,
		unsigned short length,
		unsigned char *data) {
  queue_add(q, 0xBE);
  queue_add(q, type & 0x00FF);
  queue_add(q, type >> 8);
  queue_add(q, id & 0x00FF);
  queue_add(q, id >> 8);
  queue_add(q, length & 0x00FF);
  queue_add(q, length >> 8);

  unsigned char checksum = 0;
  checksum ^= 0xBE;
  checksum ^= type & 0x00FF;
  checksum ^= type >> 8;
  checksum ^= id & 0x00FF;
  checksum ^= id >> 8;
  checksum ^= length & 0x00FF;
  checksum ^= length >> 8;
  int i;
  for (i = 0; i < length; i += 1) {
    queue_add(q, data[i]);
    checksum ^= data[i];
  }
  queue_add(q, checksum);
  return checksum;
}

// Spying into communication.c's global variables
extern u8 packet[MAX_PACKET_SIZE];
extern int bytes_recv;
extern unsigned char packet_checksum;

// Test fails when no BE and run out
int test_try_receive_packet_fails_when_no_BE_and_run_out() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(5);
  queue_add(queue, 0);
  queue_add(queue, 0);
  queue_add(queue, 1);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 0);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 0);

  return 0;
}

// Test fails when no BE and too much (check resume)
int test_try_receive_packet_fails_when_no_BE_and_too_much() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  int size = 255;
  queue = queue_malloc(size);
  queue_add_corruption(queue, size);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 0);

  // Ensure that we quit trying
  test_assert(size - queue_size(queue) <= MAX_PACKET_SIZE);

  return 0;
}

// Test fails when BE and run out (check resume)
int test_try_receive_packet_fails_when_BE_and_run_out() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(100);
  queue_add_corruption(queue, 10);
  queue_add(queue, 0xBE);

  queue_add(queue, 1);
  queue_add(queue, 2);
  queue_add(queue, 3);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 4);

  test_assert(packet[0] == 0xBE);
  test_assert(packet[1] == 1);
  test_assert(packet[2] == 2);
  test_assert(packet[3] == 3);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 4);

  return 0;
}

// Test fails when BE and nonsensical length (check resume)
int test_try_receive_packet_fails_when_BE_and_big_length() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(500);
  int i;
  queue_add_corruption(queue, 10);
  queue_add(queue, 0xBE);
  queue_add_short(queue, 0);
  queue_add_short(queue, 0);
  queue_add_short(queue, 500);
  for (i = 0; i < 10; i += 1) queue_add_corruption(queue, i);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 0);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 0);

  return 0;
}

// Test fails when BE and length and run out (check resume)
int test_try_receive_packet_fails_when_BE_length_and_run_out() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(500);
  int i;
  queue_add_corruption(queue, 20);
  queue_add(queue, 0xBE);
  queue_add_short(queue, 0);
  queue_add_short(queue, 0);
  queue_add_short(queue, 4);
  unsigned char data[4] = {1, 2, 3, 4};
  for (i = 0; i < 2; i += 1) queue_add(queue, data[i]);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 9);
  test_assert(packet[0] == 0xBE);
  test_assert(packet[1] == 0);
  test_assert(packet[2] == 0);
  test_assert(packet[3] == 0);
  test_assert(packet[4] == 0);
  test_assert(packet[5] == 4);
  test_assert(packet[6] == 0);
  test_assert(packet[7] == 1);
  test_assert(packet[8] == 2);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 9);

  return 0;
}

// Test fails when BE and length and data and run out (check resume)
int test_try_receive_packet_fails_when_BE_length_data_and_run_out() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(500);
  int i;
  queue_add_corruption(queue, 10);
  queue_add(queue, 0xBE);
  queue_add_short(queue, 0);
  queue_add_short(queue, 0);
  queue_add_short(queue, 4);
  unsigned char data[4] = {1, 2, 3, 4};
  for (i = 0; i < 4; i += 1) queue_add(queue, data[i]);

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 11);
  test_assert(packet[0] == 0xBE);
  test_assert(packet[1] == 0);
  test_assert(packet[2] == 0);
  test_assert(packet[3] == 0);
  test_assert(packet[4] == 0);
  test_assert(packet[5] == 4);
  test_assert(packet[6] == 0);
  test_assert(packet[7] == 1);
  test_assert(packet[8] == 2);
  test_assert(packet[9] == 3);
  test_assert(packet[10] == 4);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 11);

  return 0;
}

// Test fails when BE, length, data, and checksum fails
int test_try_receive_packet_fails_when_BE_length_data_and_bad_checksum() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(500);
  int i;
  queue_add_corruption(queue, 10);
  queue_add(queue, 0xBE);
  queue_add_short(queue, 0);
  queue_add_short(queue, 0);
  queue_add_short(queue, 4);
  unsigned char data[4] = {1, 2, 3, 4};
  for (i = 0; i < 4; i += 1) queue_add(queue, data[i]);
  queue_add(queue, 0xFF); // bad checksum
  queue_add(queue, 0xBE); // next start

  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 1);
  test_assert(packet[0] == 0xBE);

  // Try again to verify that we actually ran out
  test_assert(try_receive_packet(uart));
  test_assert(bytes_recv == 1);

  return 0;
}

// Test succeeds when BE, length, data, checksum
int test_try_receive_packet_succeeds() {
  bytes_recv = 0;
  uart = mock_uart_malloc();
  queue = queue_malloc(500);
  int i;
  queue_add_corruption(queue, 10);;
  unsigned char data[4] = {1, 2, 3, 4};
  unsigned char checksum = queue_add_packet(queue, 0, 0, 4, data);

  int failure = try_receive_packet(uart);
  test_assert(!failure);
  test_assert(bytes_recv == 12);
  test_assert(packet[0] == 0xBE);
  test_assert(packet[1] == 0);
  test_assert(packet[2] == 0);
  test_assert(packet[3] == 0);
  test_assert(packet[4] == 0);
  test_assert(packet[5] == 4);
  test_assert(packet[6] == 0);
  test_assert(packet[7] == 1);
  test_assert(packet[8] == 2);
  test_assert(packet[9] == 3);
  test_assert(packet[10] == 4);
  test_assert(packet[11] == checksum);

  // Try again to verify that we don't lose the ready packet
  test_assert(!try_receive_packet(uart));
  test_assert(bytes_recv == 12);

  return 0;
}

int main() {
  TEST(test_try_receive_packet_fails_when_no_BE_and_run_out);
  TEST(test_try_receive_packet_fails_when_no_BE_and_too_much);
  TEST(test_try_receive_packet_fails_when_BE_and_run_out);
  TEST(test_try_receive_packet_fails_when_BE_and_big_length);
  TEST(test_try_receive_packet_fails_when_BE_length_and_run_out);
  TEST(test_try_receive_packet_fails_when_BE_length_data_and_run_out);
  TEST(test_try_receive_packet_fails_when_BE_length_data_and_bad_checksum);
  TEST(test_try_receive_packet_succeeds);
  return test_summary();
}
